/* General Purpose DDS, for Joseph Thywissen's labs.
	Started April 2011, Alan Stummer

	DESCRIPTION
	This is the next in a series of DDSs.  All previous designs have been for specific uses, this one
	is just a DDS for generic frequency generation.  It is back compatible with existing user software.
	
	The heart of the unit is a AD9910 evaluation board.  It is controlled by an Altera Cyclone III
	FPGA.  The FPGA is controlled by a Rabbit (now Digi) module, which talks with the host computer
	over ethernet.
	
	The DDS is serial programmed by the FPGA.  This includes discrete frequencies and ramps.	The host
	sends the desired sequence to the Rabbit which in turn loads the sequence into the FPGA.  In
	normal use, a realtime computer will send a trigger pulse to the FPGA.  The FPGA gets the next
	frequency or ramp in sequence and programs it into the DDS.
	
	COMMAND SEQUENCE:  See the file "All About General Purpose DDS.txt for details.
	
	Note that no error checking on Rabbit stuff is done, it is assumed that there are no errors.  If
	errored data is sent, it will have unknown results.
	
	It is a proven scientific fact that dogs run faster in Greenland.  This is because the trees are
	farther apart.  Does anyone read this header?
	*/
	
//synthesis VERILOG_INPUT_VERSION SYSTEMVERILOG_2005
	module uberDDS (
		//with Rabbit
		input   [2:0] rSpare,																//spares, for future use.
		input   [3:0] rCmd,																  //command bus from Rabbit.
		input         rCS,																			  //CS from Rabbit.
		inout   [7:0] rData,																	  //data bus with Rabbit.
		output        CmdSeqFull = 0,							 //flag to show when command sequence is full.
		//IOs
		input         Trigger,										  //buffered TTL Trigger input from Adwin.
		input         TriggerRaw,									//unbuffered TTL Trigger input from Adwin.
		input         Spare_in1, Spare_in2,															  //spare input.
		output        Q1, Q2, Q3,																  //spare TTL outputs.
		output        TP1, TP2,																			  //test points.
		//with DDS
		input         DRover,							//goes high when ramp hits its upper or lower limit.
		output        DRctl,															  //controls direction of ramp.
		output        DRhold,																	  //high freezes ramp.
		output        SDio,											//serial data I/O, used here always as SDI.
		input         Sdo,																			 //serial data out.
		output        Sclk,																		  //serial data clock.
		output        IOrst,																		  //serial port reset.
		output        IOupdate,															  //serial data is executed.
		output  [0:2] P,																			  //profile selection.
		input         PLLlock,															  //high when PLL is locked.
		input         RAMswpOvr,											  //true when RAM sweep has finished.
		output        TXen,																		 //enables DAC output.
		output        Rst = 0,																			 //master reset.
		//misc
		input         Clk,																				 //100MHz clock.
		//with Cmd Seq RAM
		input   [7:0] CmdSeqDataRead,											 //data read from the cmd seq RAM.
		output [14:0] CmdSeqAdr,																//cmd seq RAM address.
		output        CmdSeqWrite = 0										//true allows writing to cmd seq RAM.
		);
		
	//Parameters...
	localparam       VerRev        = 8'h07;			//SET THE VERSION(MS):REVISION((LS) NUMBERS HERE!
	localparam       CmdSeqMaxSize = 15'd32750;						//max byte count of command sequence.
	localparam       START_CMDS       = 4'd0;
	localparam       CMD_SIZE         = 4'd1;
	localparam       NOT_A_REG        = 4'd2;
	localparam       WAIT_FOR_RAMP    = 4'd3;
	localparam       WAIT_FOR_TRIGGER = 4'd4;
	localparam       HDWR_PIN         = 4'd5;
	localparam       NEXT_BIT         = 4'd6;
	localparam       CLOCK_HIGH       = 4'd7;
	localparam       CLOCK_LOW        = 4'd8;
	localparam       NEXT_BYTE        = 4'd9;
	localparam       FINISHED_CMD     = 4'd10;
	localparam       RESTART_CMD_SEQ  = 4'd11;
	
	//Globals...
	reg       [14:0] CmdSeqSize = 15'd0;							  //number of bytes in command sequence.
	reg        [7:0] ByteCtr;									  //counter for number of bytes bung into DDS.
	reg        [7:0] BitMask;										  //pointer for bit banging bytes into DDS.
	reg              LastrCS = 1;						 //previous state of rCS, for rising edge detection.
	reg              RunSequence = 0;						  //true allows the sequenced commands to run.
	reg        [3:0] State;							  //state machine pointer for running command sequence.
	reg              IOupdateRun = 0;										  //IOupdate when running cmd seq.
	reg              TrigDirect = 0;					 //directly plumb Trigger input to IO_UPDATE output.

	assign P    = 3'b000;																			  //profile number.
	assign TXen = 1;												  //transmit enable, AKA enable DDS RF output.
	assign Q1   = RunSequence;
	assign Q2   = Trigger;
	assign Q3   = IOupdate;
	assign TP1  = 0;
	assign TP2  = 0;
	assign IOupdate = (RunSequence && IOupdateRun)		  //Trigger has no direct effect when running.
                 || (!RunSequence && TriggerRaw && TrigDirect); //Trigger to IOupdate asyncronously.

	
/******************************************** Rabbit Command Handler *****************************/
	always @(negedge Clk)  begin
//		if (RunSequence & (CmdSeqSize+1 > CmdSeqAdr))  Q3 <= 1;				//running command sequence.
//		else  if (RunSequence & DRover)                Q3 <= 0;
		LastrCS <= rCS;															//save rCs to detect rising edge.

		if (rCS)  begin											//rCS FROM RABBIT IS TRUE, IMPLEMENT COMMAND.
			case (rCmd)
				4'd0:  begin												//Serial Write Direct To DDS Registers.
					IOrst           <= rData[0];
					Sclk            <= rData[1];
					SDio            <= rData[2];
					Rst             <= rData[3];
					//[4] used to be I/O_UPDATE pin, removed, not needed.
					DRhold          <= rData[5];
					DRctl           <= rData[6];
				end

				4'd1:  begin																  //Reset Command Sequence.
					CmdSeqSize  <= 15'd0;
					CmdSeqAdr   <= 15'd0;							  //address 0 not used, first cmd is at 1.
					RunSequence <= 0;
				end
			
				4'd2:  begin					  //Return To Start Of Command Sequence So Can Run It Again.
					RunSequence <= 0;
					CmdSeqAdr   <= 15'd1;											  //first cmd is at address 1.
				end

				4'd3:  if (CmdSeqSize)  RunSequence <= 1;					//Start Running Command Sequence.
			
				4'd4:  begin														  //Add Byte To Command Sequence.
					if (!LastrCS & !CmdSeqFull)  begin							//rCS from Rabbit rising edge.
						CmdSeqSize <= CmdSeqSize + 15'd1;								  //inc size of cmd seq.
						CmdSeqAdr  <= CmdSeqAdr  + 15'd1;									//go to next address.
						CmdSeqWrite <= 1;
					end
					if (LastrCS & !CmdSeqFull)  CmdSeqWrite <= 0;	  //rCS from Rabbit now steady high.
					if (CmdSeqSize >= CmdSeqMaxSize)  CmdSeqFull <= 1;			//command sequence is full.
					else                              CmdSeqFull <= 0;
				end
				
				4'd5:  TrigDirect <= 1;		 //plumb Trigger input directly (asyncronously) to IOupdate.

				4'd6:  TrigDirect <= 0;		 					 //don't plumb Trigger input to IOupdate pin.

				4'd7:  rData <= VerRev;				 //Read FPGA Version (4-bit MS, 4-bit LS, big endian).
			
				default: ;
			endcase
		end
		

		else  begin									 //ALWAYS HERE EXCEPT FOR RISING EDGE OF rCS FROM RABBIT.
			CmdSeqWrite <= 0;								//if was writing a byte into cmd seq, then stop it.
			rData <= 8'dz;						 //tristate data bus when rCS goes false after reading data.
		end


	
/**************************************** Execute Command Sequence ********************************/
		if (RunSequence)  begin
			case (State)
				START_CMDS:  begin															//POINT TO FIRST COMMAND.
					CmdSeqAdr   <= 15'd1;												//first cmd is at address 1.
					IOrst       <= 0;
					Sclk        <= 0;
					Rst         <= 0;
					IOupdateRun <= 0;
					ByteCtr     <= 8'd0;
					TrigDirect  <= 0;
					State       <= CMD_SIZE;
				end
					
				CMD_SIZE:  begin															 //GET SIZE OF NEXT COMMAND.
					if (CmdSeqDataRead == 8'd1)  State <= NOT_A_REG;
					else  begin													  //It's A Register Address And Data.
						ByteCtr <= CmdSeqDataRead;			 //number of bytes in this command (addr + data).
						BitMask <= 8'b10000000;								  //point to MSB of register address.
						State   <= NEXT_BIT;
					end
					CmdSeqAdr <= CmdSeqAdr + 15'd1;												//point to command.
				end

				NOT_A_REG:  begin														  //NOT A DDS REGISTER COMMAND.
					case (CmdSeqDataRead)
						8'hA4:    State <= WAIT_FOR_TRIGGER;
						8'hA8:    State <= WAIT_FOR_RAMP;
						8'hC4:    State <= RESTART_CMD_SEQ;
						default:  State <= HDWR_PIN;
					endcase
				end

				WAIT_FOR_RAMP:  begin														  //WAIT FOR RAMP TO END.
					if (DRover)  begin
						CmdSeqAdr <= CmdSeqAdr + 15'd1;									 //point to next command.
						State     <= FINISHED_CMD;
					end
				end

				RESTART_CMD_SEQ:  begin														  //RESTART THE CMD SEQ.
					CmdSeqAdr <= 15'd1;									  //point to first command.
					State     <= FINISHED_CMD;
				end
				
				WAIT_FOR_TRIGGER:  begin															//WAIT FOR TRIGGER.
					if (Trigger)  begin
						State <= FINISHED_CMD;
						CmdSeqAdr <= CmdSeqAdr + 15'd1;									 //point to next command.
					end
				end

				HDWR_PIN:  begin																		  //SET A HDWR PIN.
					case (CmdSeqDataRead)
						8'h01:  IOrst       <= 0;											  //set IO Reset pin low.
						8'h02:  IOrst       <= 1;											 //set IO Reset pin high.
						8'h03:  IOupdateRun <= 0;											 //set IO Update pin low.
						8'h04:  IOupdateRun <= 1;											//set IO Update pin high.
						8'h05:  DRhold      <= 0;											   //set DR Hold pin low.
						8'h06:  DRhold      <= 1;											  //set DR Hold pin high.
						8'h07:  DRctl       <= 0;											   //set DR Ctrl pin low.
						8'h08:  DRctl       <= 1;											  //set DR Ctrl pin high.
						default:  ;
					endcase
					CmdSeqAdr <= CmdSeqAdr + 15'd1;										 //point to next command.
					State <= FINISHED_CMD;											//execute next command, if any.
				end
				
				NEXT_BIT:  begin															  //SET ADDRESS OR DATA BIT.
					if (CmdSeqDataRead & BitMask)  SDio <= 1;										  //set the bit.
					else                           SDio <= 0;										//clear the bit.
					BitMask <= {1'b0, BitMask[7:1]};		 		 //point to next bit (using concatenation).
					State   <= CLOCK_HIGH;
				end
				
				CLOCK_HIGH:  begin																	  //SET CLOCK HIGH.
					Sclk  <= 1;
					State <= CLOCK_LOW;
				end
				
				CLOCK_LOW:  begin																			//SET CLOCK LOW.
					Sclk  <= 0;
					if  (BitMask)  State <= NEXT_BIT;						  //bang in next bit in this byte.
					else           State <= NEXT_BYTE;											 //go to next byte.
				end
				
				NEXT_BYTE:  begin														//GO TO NEXT BYTE, OR FINISHED.
					CmdSeqAdr <= CmdSeqAdr + 15'd1;											 //point to next byte.
					if (ByteCtr > 8'd1)  begin														 //Go To Next Byte.
						ByteCtr <= ByteCtr - 8'd1;									 //another byte bites the dust.
						BitMask <= 8'b10000000;															 //point to MSB.
						State   <= NEXT_BIT;															//start loading it.
					end
					else  State <= FINISHED_CMD;								 //finished loading this register.
				end
				
				FINISHED_CMD:  begin					//FINISHED THIS COMMAND SEQUENCE, WAIT FOR LOW TRIGGER.
					if (!Trigger & (CmdSeqSize+1 > CmdSeqAdr))  State <= CMD_SIZE;		//execute next cmd.
					else  ;		//if that was last command, spin here forever until RunSequence is false.
				end
				
				default: ;
			endcase
		end
		
		
		else  State <= START_CMDS;													//NOT RUNNING COMMAND SEQUENCE.
	end


endmodule
